import traceback


def extract_error(e: Exception, pattern='%s: %s', *, raw=False) -> tuple[str, str] | str:
  if callable(e):
    e = e()
  one = e.__class__.__name__
  two = str(e)
  return (one, two) if raw else pattern % (one, two)


def extract_traceback(e: Exception) -> str:
  traceback_details = traceback.format_tb(e.__traceback__)
  return ''.join(traceback_details)



from typing import Any, Union


class Null:
  def __str__(self):
    return self.__repr__()

  def __repr__(self):
    return 'Null'


def or_(self, __value: Any) -> Union[Null, Any]:  # noqa: UP007
  return Union[Null, __value]  # noqa: UP007


def ror_(self, __value: Any) -> Union[Null, Any]:  # noqa: UP007
  return Union[Null, __value]  # noqa: UP007


def _new(cls, *args, **kwargs):
  return Null


Null.__or__ = or_
Null.__ror__ = ror_

Null, NullType = Null(), Null

NullType.__new__ = _new
Null.__class__ = NullType

assert (Null) is (Null.__class__())  # No sneaky sneakies with making two different Nulls


class UniqueDefault:
  """Used as a default parameter for the parameter named default (yeah i know that's uh... let's just see the example).

  Example:
  ```py
  from tcrutils import UniqueDefault as RaiseError

  def get_something(..., *, default: Any | RaiseError = RaiseError):
  ... # Function implementation

  # Nothing was found, falling to raising error or returning default value if such was specified.
  if default is not RaiseError:
    return default
  raise KeyError("Unable to find something.")
  """



from .tcrd_constants import DISCORD_EPOCH
from .tcrd_embeds import embed
from .tcrd_limits import DiscordLimits
from .tcrd_permissions import PERMISSIONS_DICT, Permission, permissions
from .tcrd_permissions import Permission as DiscordPermission
from .tcrd_permissions import permissions as discord_permissions
from .tcrd_snowflake import Snowflake, is_snowflake
from .tcrd_string import get_token

__all__ = [
  'DiscordLimits',
  'DiscordPermission',
  'discord_permissions',
  'tcrd_embeds',
  "is_snowflake", "Snowflake",
  "embed",
  'get_token',
]



def commafy(text: str | int, splitter: str = ','):
  if isinstance(text, int) and text < 0:
    return '-' + commafy(-text)
  text = str(text)
  temp = ''
  for i, letter in enumerate(text[::-1]):
    temp += letter
    if i % 3 == 2 and i != len(text) - 1:
      temp += splitter
  return temp[::-1]



import typing as t

from colored import attr, bg, fg, stylize

from .tcr_extract_error import extract_error

color_aliases = {
  'gold': 'yellow',
}


class Color:
  RESET: str  # Reset constant
  _default: str | None  # Color that will be applied on invalid color

  def __init__(self) -> None:
    self.RESET = stylize('', attr(0))
    self._default = None

  def _get_default_color(self, name=''):
    if self._default is None:
      msg = f'Invalid color: {name!r}'
      raise ValueError(msg)
    return self._default

  @property
  def default(self):
    return self._default

  @default.setter
  def default(self, color: str | None = None) -> str:
    if color is None:
      self._default = None
      return
    if not isinstance(color, str):
      msg = f'Invalid color type, must be str, got {type(color)}'
      raise TypeError(msg)
    self._default = self(color)

  def __call__(self, *colors: str | None) -> str:
    """Abstraction for easier ansi coloring & formatting."""

    if len(colors) >= 2:
      return ''.join([self.__call__(x) for x in colors])
    else:
      if colors == ():
        return self.RESET
      colors = colors[0]

    if str(colors).lower() in ['reset', '0']:
      return self.RESET

    if isinstance(colors, int):
      return fg(colors)

    colors = colors.replace(
      '\\_', '🚜🚛🚚🚚🚚🚒🚝🚈🚈🚈🚅🚅🚄🚔🚔🚘🚘🚜🚜🚛🚛🚛🚚🚚🚒🚒🚑🚑'
    )  # Underlined if it contains _
    attr_und = '_' in colors
    colors = colors.replace('_', '')  # Underlined if it contains _
    colors = colors.replace(
      '🚜🚛🚚🚚🚚🚒🚝🚈🚈🚈🚅🚅🚄🚔🚔🚘🚘🚜🚜🚛🚛🚛🚚🚚🚒🚒🚑🚑', '_'
    )  # Underlined if it contains _
    attr_dim = '&' in colors
    colors = colors.replace('&', '')  # Dim        if it contains &
    attr_bli = '*' in colors
    colors = colors.replace('*', '')  # Blink      if it contains *
    attr_rev = '!' in colors
    colors = colors.replace('!', '')  # Reverse    if it contains !
    attr_hidd = '#' in colors
    colors = colors.replace('#', '')  # Hidden     if it contains #

    attr_bold = colors[0] == colors[0].upper()  # Bold if the first letter is uppercase
    is_bg = colors == colors.upper() and not colors.isnumeric()  # bg if all caps

    colors = colors.lower()

    if colors in color_aliases:
      colors = color_aliases[colors]

    styles = []
    try:
      styles.append((bg if is_bg else fg)(colors))
    except KeyError:
      return self._get_default_color(colors)
    if attr_bold:
      styles.append(attr('bold'))
    if attr_und:
      styles.append(attr('underlined'))
    if attr_dim:
      styles.append(attr('dim'))
    if attr_bli:
      styles.append(attr('blink'))
    if attr_rev:
      styles.append(attr('reverse'))
    if attr_hidd:
      styles.append(attr('hidden'))

    return stylize('', ''.join(styles), reset=False)

  def __str__(self) -> str:
    return c()


color = Color()
del Color
c, colour = color, color


def printc(*args, end='\n', **kwargs):
  end = c('reset') + end
  return print(*args, end=end, **kwargs)



from collections.abc import Callable

from ..discord.tcrd_limits import DiscordLimits
from .tcr_constants import BACKTICKS, NEWLINE
from .tcr_iterable import cut_at


def codeblock(
  text: str,
  *,
  langcode='',
  max_length: int = DiscordLimits.Message.LENGTH_SAFE,
  cut_at_func: Callable[[str, int], str] = cut_at,
) -> str:
  """Pack text into a Discord codeblock.

  May return str with len > max_length if max_length is ridicously low such that not even `'''\\nlangcode'''` fits
  """
  maxlen = (
    max_length - (len(BACKTICKS) + len(langcode) + len(NEWLINE) + len(BACKTICKS))
    if max_length != -1
    else 9999999999999999  # Eh...
  )
  return BACKTICKS + langcode + NEWLINE + cut_at_func(text, maxlen) + BACKTICKS


def uncodeblock(text: str) -> str:
  """*Try to* remove any discord codeblocks along with their langcodes if any are found."""
  if text[-3:] == BACKTICKS and text[:3] == BACKTICKS:
    code_start = 3
    code_end = -3
    if NEWLINE in text[3:]:  # Check if there is a language code specified
      code_start = text.index(NEWLINE) + 1
    return text[code_start:code_end].strip()
  return text



from collections.abc import Generator, Hashable, Mapping
from typing import Any


def _check_strict(master: Mapping, slave: Mapping, *, recursive: bool) -> bool:
  for k, v in master.items():
    if k not in slave:
      return [k]
    if recursive and isinstance(v, Mapping) and isinstance(slave[k], Mapping):  # noqa: SIM102
      if (a := _check_strict(v, slave[k], recursive=recursive)) is not True:
        return [k, *a]
  return True


def dict_zip(dict1: Mapping, *dicts: Mapping) -> Generator[tuple[Hashable, Any], None, None]:
  """Zips dictionaries simillar to `zip()`, returns a generator which yields a tuple of (key, (value1, value2, value3)) for each key. This function does not support analogy of `zip()` non-strict mode, strict=True is implied (Each key must be in every single passed in dict)."""

  dicks = [dict1, *dicts]

  kys = dicks[0].keys()
  if any(kys != dick.keys() for dick in dicks):
    msg = 'Mismatched keys'
    raise ValueError(msg)

  for k, v in dicks[0].items():
    yield (k, (v, *(dick[k] for dick in dicks[1:])))


def merge_dicts(
  master: Mapping, slave: Mapping, *dicts: Mapping, recursive=True, strict=False
) -> Mapping:
  """Merge dictionaries, made to prioritize the `master` dictionary and if key is not found there, then it takes from the `slave` dictionary.

  Optionally if `recursive=True`, then if the same key is a dict in both master and slave, merge dicts operation is performed on both of them.\\
  Optionally if `strict=True` it raises `ValueError` if there exists a key that is in master dictionary but not in slave dictionary.\\
  If there are more than 2 arguments passed in the list is reduced from back to front until two elements are left by the process of merge_dicts(master=list[-2], slave=list[-1])
  """
  merged = {}

  dicts = [master, slave, *dicts]

  while len(dicts) > 2:
    slave_ = dicts.pop()
    master_ = dicts[-1]
    dicts[-1] = merge_dicts(master_, slave_, recursive=recursive, strict=strict)

  master, slave = dicts

  if strict:
    a = _check_strict(master, slave, recursive=recursive)
    if a is not True:
      msg = f'Strict check failed: there exists a key ({".".join(a)}) that is in master dictionary but not in slave dictionary'
      raise ValueError(msg)

  for key in master:
    if key in slave:
      if isinstance(master[key], dict) and isinstance(slave[key], dict) and recursive:
        merged[key] = merge_dicts(master[key], slave[key], recursive=True)
      else:
        merged[key] = master[key]
    else:
      merged[key] = master[key]

  for key in slave:
    if key not in merged:
      merged[key] = slave[key]

  return merged



import os
import pathlib as p
import re as regex
from typing import Literal, NoReturn

from ..src.tcr_compare import able
from ..src.tcr_constants import BACKTICKS
from ..src.tcr_null import UniqueDefault as RaiseError
from .tcrd_snowflake import is_snowflake


def get_token(
  filename: str = 'TOKEN.txt', depth=2, *, dont_strip=False, default=RaiseError
) -> None | str:
  """Get the nearest file with name=filename (default 'TOKEN.txt') and return its stripped contents (unless specified not to strip with `dont_strip=True`).

  This algoritm searches for files named TOKEN.txt (or custom name) in the current directory, then the parent directory, then the parent of parent and so on.
  The search continues up to depth `depth` (depth=0: current directory only, depth=1, current directory and its parent, etc.). If multiple files are found, return the closest one to the current directory.
  If no file is found, an error will be raised unless `default` is provided.
  """

  def rexit(x):
    os.chdir(origin_path)
    return x

  origin_path = p.Path.cwd().absolute()

  os.chdir(origin_path)

  f = origin_path / filename

  for i in range(depth + 1):
    f = (origin_path / '/'.join(['..'] * i)) / filename
    if f.is_file():
      t = f.read_text()
      if not dont_strip:
        t = t.strip()
      return rexit(t)

  if default is not RaiseError:
    return default

  raise FileNotFoundError(f'Unable to locate token file: {filename}')


class IFYs:
  """Features related to turning IDs into user/channel/command/etc. mentions, emojis, and more if i can think of any.

  Example:
  ```py
  >>> IFYs.userify(1234)
  '<@1234>'
  ```

  Includes argument validation:
  - Raises TypeError if invalid type was supplied (not str or int)
  - Raises ValueError if the type is valid, but snowflake in it is not (for example if negative int was passed in)
  """

  @staticmethod
  def userify(user_id: int | str):
    """### User mentions.

    ```py
    >>> IFYs.userify(1234)
    '<@1234>'
    ```
    """

    if not is_snowflake(user_id, allow_string=True):
      if isinstance(user_id, int | str):
        err = ValueError(
          'user_id is not a valid snowflake, use tcr.discord.is_snowflake() to validate if needed.'
        )
      else:
        err = TypeError(f'Expected str or int, got {type(user_id)} instead.')

      raise err

    return f'<@{user_id}>'

  @staticmethod
  def channelify(channel_id: int | str):
    """### User mentions.

    ```py
    >>> IFYs.channelify(1234)
    '<#1234>'
    ```
    """

    if not is_snowflake(channel_id, allow_string=True):
      if isinstance(channel_id, int | str):
        err = ValueError(
          'channel_id is not a valid snowflake, use tcr.discord.is_snowflake() to validate if needed.'
        )
      else:
        err = TypeError(f'Expected str or int, got {type(channel_id)} instead.')

      raise err

    return f'<#{channel_id}>'
